"use strict";
/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `MSSqlDriver` and related types declaration.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MSSqlDriver = void 0;
const mssql_1 = __importStar(require("mssql"));
const shared_1 = require("@cubejs-backend/shared");
const base_driver_1 = require("@cubejs-backend/base-driver");
const QueryStream_1 = require("./QueryStream");
// ********* Value converters ***************** //
const numericTypes = [
    mssql_1.default.TYPES.Int,
    mssql_1.default.TYPES.BigInt,
    mssql_1.default.TYPES.SmallInt,
    mssql_1.default.TYPES.TinyInt,
    mssql_1.default.TYPES.Decimal,
    mssql_1.default.TYPES.Numeric,
    mssql_1.default.TYPES.Float,
    mssql_1.default.TYPES.Real,
    mssql_1.default.TYPES.Money,
    mssql_1.default.TYPES.SmallMoney
];
for (const type of numericTypes) {
    mssql_1.default.valueHandler.set(type, (value) => (value != null ? String(value) : value));
}
const GenericTypeToMSSql = {
    boolean: 'bit',
    string: 'nvarchar(max)',
    text: 'nvarchar(max)',
    timestamp: 'datetime2',
    uuid: 'uniqueidentifier'
};
const MSSqlToGenericType = {
    bit: 'boolean',
    uniqueidentifier: 'uuid',
    datetime2: 'timestamp'
};
/**
 * MS SQL driver class.
 */
class MSSqlDriver extends base_driver_1.BaseDriver {
    connectionPool;
    initialConnectPromise;
    config;
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency() {
        return 2;
    }
    /**
     * Class constructor.
     */
    constructor(config = {}) {
        super({
            testConnectionTimeout: config.testConnectionTimeout,
        });
        const dataSource = config.dataSource ||
            (0, shared_1.assertDataSource)('default');
        /**
         * @type {import('mssql').config}
         */
        this.config = {
            readOnly: true,
            server: (0, shared_1.getEnv)('dbHost', { dataSource }),
            database: (0, shared_1.getEnv)('dbName', { dataSource }),
            port: (0, shared_1.getEnv)('dbPort', { dataSource }),
            user: (0, shared_1.getEnv)('dbUser', { dataSource }),
            password: (0, shared_1.getEnv)('dbPass', { dataSource }),
            domain: (0, shared_1.getEnv)('dbDomain', { dataSource }),
            requestTimeout: (0, shared_1.getEnv)('dbQueryTimeout') * 1000,
            options: {
                encrypt: (0, shared_1.getEnv)('dbSsl', { dataSource }),
                useUTC: false
            },
            pool: {
                max: config.maxPoolSize ||
                    (0, shared_1.getEnv)('dbMaxPoolSize', { dataSource }) ||
                    8,
                min: 0,
                idleTimeoutMillis: 30 * 1000,
                acquireTimeoutMillis: 20 * 1000
            },
            ...config
        };
        const { readOnly, ...poolConfig } = this.config;
        this.connectionPool = new mssql_1.ConnectionPool(poolConfig);
        this.initialConnectPromise = this.connectionPool.connect();
    }
    /**
     * Returns the configurable driver options
     * Note: It returns the unprefixed option names.
     * In case of using multi sources options need to be prefixed manually.
     */
    static driverEnvVariables() {
        return [
            'CUBEJS_DB_HOST',
            'CUBEJS_DB_NAME',
            'CUBEJS_DB_PORT',
            'CUBEJS_DB_USER',
            'CUBEJS_DB_PASS',
            'CUBEJS_DB_DOMAIN',
        ];
    }
    async testConnection() {
        const conn = await this.initialConnectPromise.then((pool) => pool.request());
        await conn.query('SELECT 1 as number');
    }
    /**
     * Executes query in streaming mode.
     *
     * @param {string} query
     * @param {Array} values
     * @param {{ highWaterMark: number? }} options
     * @return {Promise<StreamTableDataWithTypes>}
     */
    async stream(query, values, { highWaterMark }) {
        const pool = await this.initialConnectPromise;
        const request = pool.request();
        request.stream = true;
        (values || []).forEach((v, i) => {
            request.input(`_${i + 1}`, v);
        });
        request.query(query);
        const stream = new QueryStream_1.QueryStream(request, highWaterMark);
        const fields = await new Promise((resolve, reject) => {
            request.on('recordset', (columns) => {
                resolve(this.mapFields(columns));
            });
            request.on('error', (err) => {
                reject(err);
            });
            stream.on('error', (err) => {
                reject(err);
            });
        });
        return {
            rowStream: stream,
            types: fields,
            release: async () => {
                request.cancel();
            },
        };
    }
    /**
     * @param {{
     *   [name: string]: {
     *     index: number,
     *     name: string,
     *     type: *,
     *     nullable: boolean,
     *     caseSensitive: boolean,
     *     identity: boolean,
     *     readOnly: boolean,
     *     length: number?,
     *     scale: number?,
     *     precision: number?
     *   }
     * }} fields
     */
    mapFields(fields) {
        return Object.keys(fields).map((field) => {
            let type;
            switch (fields[field].type) {
                case mssql_1.default.Bit:
                    type = 'boolean';
                    break;
                // integers
                case mssql_1.default.Int:
                case mssql_1.default.SmallInt:
                case mssql_1.default.TinyInt:
                case mssql_1.default.BigInt:
                    type = 'int';
                    break;
                // float
                case mssql_1.default.Money:
                case mssql_1.default.SmallMoney:
                case mssql_1.default.Numeric:
                case mssql_1.default.Decimal:
                    type = 'decimal';
                    break;
                // double
                case mssql_1.default.Real:
                case mssql_1.default.Float:
                    type = 'double';
                    break;
                // strings
                case mssql_1.default.Char:
                case mssql_1.default.NChar:
                case mssql_1.default.Text:
                case mssql_1.default.NText:
                case mssql_1.default.VarChar:
                case mssql_1.default.NVarChar:
                case mssql_1.default.Xml:
                    type = 'text';
                    break;
                // date and time
                case mssql_1.default.Time:
                    type = 'time';
                    break;
                case mssql_1.default.Date:
                    type = 'timestamp';
                    break;
                case mssql_1.default.DateTime:
                case mssql_1.default.DateTime2:
                case mssql_1.default.SmallDateTime:
                case mssql_1.default.DateTimeOffset:
                    type = 'timestamp';
                    break;
                // others
                case mssql_1.default.UniqueIdentifier:
                case mssql_1.default.Variant:
                case mssql_1.default.Binary:
                case mssql_1.default.VarBinary:
                case mssql_1.default.Image:
                case mssql_1.default.UDT:
                case mssql_1.default.Geography:
                case mssql_1.default.Geometry:
                case mssql_1.default.TVP:
                    type = 'string';
                    break;
                // unknown
                default:
                    type = 'string';
                    break;
            }
            return { name: fields[field].name, type: this.toGenericType(type) };
        });
    }
    async query(query, values) {
        let cancelFn = null;
        const promise = this.initialConnectPromise.then((pool) => {
            const request = pool.request();
            (values || []).forEach((v, i) => request.input(`_${i + 1}`, v));
            // TODO time zone UTC set in driver ?
            cancelFn = () => request.cancel();
            return request.query(query).then(res => res.recordset);
        });
        promise.cancel = () => cancelFn && cancelFn();
        return promise;
    }
    param(paramIndex) {
        return `@_${paramIndex + 1}`;
    }
    async tableColumnTypes(table) {
        const [schema, name] = table.split('.');
        const columns = await this.query(`SELECT column_name as ${this.quoteIdentifier('column_name')},
             table_name as ${this.quoteIdentifier('table_name')},
             table_schema as ${this.quoteIdentifier('table_schema')},
             data_type  as ${this.quoteIdentifier('data_type')}
      FROM INFORMATION_SCHEMA.COLUMNS
      WHERE table_name = ${this.param(0)} AND table_schema = ${this.param(1)}`, [name, schema]);
        return columns.map(c => ({ name: c.column_name, type: this.toGenericType(c.data_type) }));
    }
    getTablesQuery(schemaName) {
        return this.query(`SELECT table_name FROM INFORMATION_SCHEMA.TABLES WHERE table_schema = ${this.param(0)}`, [schemaName]);
    }
    async createSchemaIfNotExists(schemaName) {
        return this.query(`SELECT schema_name FROM INFORMATION_SCHEMA.SCHEMATA WHERE schema_name = ${this.param(0)}`, [schemaName]).then((schemas) => {
            if (schemas.length === 0) {
                return this.query(`CREATE SCHEMA ${schemaName}`, []);
            }
            return null;
        });
    }
    informationSchemaQuery() {
        // fix The multipart identifier "columns.data_type" could not be bound
        return `
      SELECT column_name as ${this.quoteIdentifier('column_name')},
        table_name as ${this.quoteIdentifier('table_name')},
        table_schema as ${this.quoteIdentifier('table_schema')},
        data_type as ${this.quoteIdentifier('data_type')}
      FROM INFORMATION_SCHEMA.COLUMNS
      WHERE table_schema NOT IN ('information_schema', 'sys')
    `;
    }
    async downloadQueryResults(query, values, options) {
        if (options?.streamImport) {
            return this.stream(query, values, options);
        }
        const result = await this.query(query, values);
        const types = Object.keys(result.columns).map((key) => ({
            name: result.columns[key].name,
            type: this.toGenericType(result.columns[key].type.declaration),
        }));
        return {
            rows: result,
            types,
        };
    }
    fromGenericType(columnType) {
        return GenericTypeToMSSql[columnType] || super.fromGenericType(columnType);
    }
    toGenericType(columnType) {
        return MSSqlToGenericType[columnType] || super.toGenericType(columnType);
    }
    readOnly() {
        return !!this.config.readOnly;
    }
    wrapQueryWithLimit(query) {
        query.query = `SELECT TOP ${query.limit} * FROM (${query.query}) AS t`;
    }
    capabilities() {
        return {
            incrementalSchemaLoading: true,
        };
    }
}
exports.MSSqlDriver = MSSqlDriver;
//# sourceMappingURL=MSSqlDriver.js.map